import xml.etree.ElementTree as ET
import zlib

import click
from requests.exceptions import HTTPError

from .utils import MultiHash, get_url, output_result, urljoin

NS = {
    "common": "http://linux.duke.edu/metadata/common",
    "repo": "http://linux.duke.edu/metadata/repo",
    "rpm": "http://linux.duke.edu/metadata/rpm",
}
CHUNK_SIZE = 2 * 1024 * 1024


def check_yum_repo(url: str) -> bool:
    """Validate a yum repo at url."""
    click.echo(f"Validating yum repo at {url}...")
    errors = []
    proc_packages = 0

    try:
        response = get_url(urljoin(url, "/repodata/repomd.xml"))
        repomd = ET.fromstring(response.text)
        primary_loc = repomd.find("repo:data[@type='primary']/repo:location", namespaces=NS)
        primary_url = urljoin(url, primary_loc.get("href"))

        response = get_url(primary_url)
        primary_xml = zlib.decompress(response.content, 16 + zlib.MAX_WBITS)
        primary = ET.fromstring(primary_xml)

        packages = primary.findall("common:package", namespaces=NS)
        package_count = len(packages)
        click.echo(f"Found {package_count} packages.")

        with click.progressbar(
            packages,
            label="Checking package(s)",
        ) as bar:
            for package in bar:
                location = package.find("common:location", namespaces=NS).get("href")
                checksum = package.find("common:checksum", namespaces=NS)
                package_url = urljoin(url, location)
                digest = checksum.text
                checksum_type = checksum.get("type")

                if checksum_type == "sha":
                    checksum_type = "sha1"
                multihash = MultiHash([checksum_type])

                try:
                    response = get_url(f"{package_url}", stream=True)
                except HTTPError as e:
                    errors.append(f"Could not access package at {e.response.url}: {e}")
                    continue

                for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                    multihash.update(chunk)

                if multihash.hexdigest(checksum_type) != digest:
                    errors.append(
                        f"{checksum_type} checksum mismatch for '{package_url}'. Expected "
                        f"'{digest}' but received '{multihash.hexdigest(checksum_type)}'."
                    )
                proc_packages += 1
    except HTTPError as e:
        click.echo(f"Error when attempting to access {e.response.url}: {e}", err=True)
        return False
    except KeyboardInterrupt:
        output_result(proc_packages, errors)
        raise

    return output_result(proc_packages, errors)



import re
import zlib
from typing import Optional, Set

import click
from debian.deb822 import Packages, Release
from requests.exceptions import HTTPError

from .utils import MultiHash, get_url, output_result, urljoin

CHECKSUMS = {
    "MD5sum": "md5",
    "SHA1": "sha1",
    "SHA256": "sha256",
    "SHA512": "sha512",
}
CHUNK_SIZE = 2 * 1024 * 1024


def _find_dists(url: str) -> Set[str]:
    try:
        resp = get_url(urljoin(url, "dists"))
    except HTTPError as e:
        if e.response.status_code == 404:
            raise click.ClickException(f"Could not determine dists from {url}."
                                       "Please manually supply them with --dist.")
        raise

    links = re.findall(r"href=[\"'](.*)[\"']", resp.text)
    return {dist.strip("/") for dist in links if ".." not in dist}


def _packages_file(base_url: str) -> str:
    try:
        resp = get_url(urljoin(base_url, "Packages"))
        return resp.text
    except HTTPError as e:
        if e.response.status_code == 404:
            resp = get_url(urljoin(base_url, "Packages.gz"))
            return zlib.decompress(resp.content, 16 + zlib.MAX_WBITS).decode()
        else:
            raise e


def check_apt_repo(url: str, dists: Optional[Set[str]]) -> bool:
    """Validate an apt repo."""
    click.echo(f"Validating apt repo at {url}...")
    errors = []
    proc_packages = 0

    if not dists:
        dists = _find_dists(url)
    click.echo(f"Checking dists: {', '.join(dists)}")

    try:
        for dist in dists:
            dist_url = urljoin(url, "dists", dist)
            release = get_url(urljoin(dist_url, "Release")).text
            release_file = Release(release)

            components = release_file["Components"].split()
            architectures = release_file["Architectures"].split()

            for comp in components:
                for arch in architectures:
                    packages = _packages_file(urljoin(dist_url, comp, f"binary-{arch}"))
                    # count the paragraphs
                    package_count = len(re.findall(r"^Package: ", packages, re.MULTILINE))
                    click.echo(f"Checking {dist}/{comp}/{arch}. Found {package_count} package(s).")

                    with click.progressbar(
                        Packages.iter_paragraphs(packages, use_apt_pkg=False),
                        label="Checking packages(s)",
                        length=package_count,
                    ) as bar:
                        for package in bar:
                            checksums = set(CHECKSUMS.keys()) & set(package.keys())
                            checksum_types = {key: CHECKSUMS[key] for key in checksums}
                            multihash = MultiHash(list(checksum_types.values()))
                            file_url = urljoin(url, package["Filename"])

                            try:
                                response = get_url(f"{file_url}", stream=True)
                            except HTTPError as e:
                                errors.append(
                                    f"Could not access package at {e.response.url}: {e}"
                                )
                                continue

                            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                                multihash.update(chunk)

                            for key, alg in checksum_types.items():
                                if multihash.hexdigest(alg) != package[key]:
                                    errors.append(
                                        f"{key} checksum mismatch for '{package['Filename']}'. "
                                        f"Expected '{package[key]}' but received "
                                        f"'{multihash.hexdigest(alg)}'."
                                    )
                            proc_packages += 1
    except HTTPError as e:
        click.echo(f"Error when attempting to access {e.response.url}: {e}", err=True)
        return False
    except KeyboardInterrupt:
        output_result(proc_packages, errors)
        raise

    return output_result(proc_packages, errors)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['repoaudit']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'python-debian>=0.1.44,<0.2.0',
 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['repoaudit = repoaudit:main']}

setup_kwargs = {
    'name': 'repoaudit',
    'version': '0.1.0',
    'description': 'CLI to validate yum/apt repositories',
    'long_description': "# repoaudit\n\nA tool for validating apt and yum repositories.\n\n## Installation\n\nFirst install poetry:\n\nhttps://python-poetry.org/docs/#installation\n\nThen run `poetry install` to install repoaudit's dependencies.\n\n### Development\n\nTo load the poetry shell:\n\n```\npoetry shell\nrepoaudit\n```\n\nAltenatively you can run:\n\n```\npoetry run repoaudit\n```\n\n## Usage\n\nTo get a list of commands and options:\n\n```\nrepoaudit --help\n```\n\n### Examples\n\n```\n# validate all distros of azure-cli apt repo\nrepoaudit apt https://packages.microsoft.com/repos/azure-cli/\n\n# validate only focal and bionic distros of azure-cli apt repo\nrepoaudit apt --dist focal --dist bionic https://packages.microsoft.com/repos/azure-cli/\n\n# validate azurecore repo\nrepoaudit yum https://packages.microsoft.com/yumrepos/azurecore/\n\n# validate all nested yumrepos\nrepoaudit yum -r https://packages.microsoft.com/yumrepos/\n```\n",
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import re
from typing import List

import click
from requests.exceptions import HTTPError

from .apt import check_apt_repo
from .utils import get_url
from .yum import check_yum_repo

recursive_option = click.option(
    "--recursive",
    "-r",
    is_flag=True,
    help=(
        "Attempt to recursively check repos. Requires URL to point to a directory "
        "listing with links to the repos."
    ),
)


def _get_repo_urls(url: str) -> List[str]:
    try:
        resp = get_url(url)
    except HTTPError as e:
        raise click.ClickException(
            f"{e}\n"
            "Please check the url or explicitly use repo urls without the --recursive option."
        )
    links = re.findall(r"href=[\"'](.*)[\"']", resp.text)
    return [f"{url}/{link}" for link in links if ".." not in link]


@click.group()
def main() -> None:
    """Audit a repo by validating its repo metadata and packages."""
    pass


@main.command()
@recursive_option
@click.argument("url")
@click.option("--dists", help="Comma separated list of distributions.")
def apt(recursive: bool, url: str, dists: str) -> None:
    """Validate an apt repository at URL."""
    if recursive:
        urls = _get_repo_urls(url)
    else:
        urls = [url]

    if dists:
        dist_set = set(dists.split(","))
    else:
        dist_set = None

    for repo_url in urls:
        check_apt_repo(repo_url, dist_set)


@main.command()
@recursive_option
@click.argument("url")
def yum(recursive: bool, url: str) -> None:
    """Validate a yum repository at URL."""
    if recursive:
        urls = _get_repo_urls(url)
    else:
        urls = [url]

    for repo_url in urls:
        check_yum_repo(repo_url)




import hashlib
from typing import List, Optional

import click
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


class MultiHash:
    """A class to can handle multiple hash objs."""

    def __init__(self, algorithms: List[str]) -> None:
        """Define hashers with list of algorithms."""
        self.hashers = {}
        for alg in algorithms:
            self.hashers[alg] = hashlib.new(alg)

    def update(self, chunk: bytes) -> None:
        """Update hashers with a chunk."""
        for h in self.hashers.values():
            h.update(chunk)

    def hexdigest(self, alg: str) -> str:
        """Return the hasher hex digest for algorithm."""
        return self.hashers[alg].hexdigest()


def output_result(proc_packages: int, errors: List[str]) -> bool:
    """Output number of packages processed and errors."""
    click.echo(f"Checked {proc_packages} package(s).")
    if errors:
        click.echo(f"{len(errors)} errors found:\n" + ("\n").join(errors))
        return False
    else:
        click.echo("No errors detected.")
        return True


def urljoin(*paths: str) -> str:
    """Join together a set of url components."""
    # urllib's urljoin has a few gotchas and doesn't handle multiple paths
    return "/".join(map(lambda path: path.strip("/"), paths))


def retry_session(retries: int = 3) -> requests.Session:
    """Create a requests.Session with retries."""
    session = requests.Session()
    retry = Retry(total=retries)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def get_url(
    url: str,
    stream: bool = False,
    session: Optional[requests.Session] = None
) -> requests.Response:
    """Call requests.get() on a url and return the requests.Response."""
    if not session:
        session = retry_session()
    resp = session.get(url, stream=stream)
    resp.raise_for_status()
    return resp
